#version 430
#extension GL_ARB_separate_shader_objects : enable










struct SComputeInput
{
	uvec3 GlobalThreadID; uvec3 LocalThreadID; uvec3 GroupID;};

layout(std430 ) readonly buffer BStreamInfo
{
	uint StreamInfo[];
} StreamInfo;
layout(std430 ) readonly buffer BGPUSimData
{
	uint GPUSimData[];
} GPUSimData;
layout(std430 ) readonly buffer BPositionsOffsets
{
	uint PositionsOffsets[];
} PositionsOffsets;
layout(std430 ) readonly buffer BScalesOffsets
{
	uint ScalesOffsets[];
} ScalesOffsets;
layout(std430 ) readonly buffer BOrientationsOffsets
{
	uint OrientationsOffsets[];
} OrientationsOffsets;
layout(std430 ) readonly buffer BMatricesOffsets
{
	uint MatricesOffsets[];
} MatricesOffsets;
layout(std430 ) buffer BOutMatrices
{
	uint OutMatrices[];
} OutMatrices;

layout( std140) uniform UInfo
{
	uint DrawRequest;
} Info;

void ComputeMain( in SComputeInput  cInput  )
{
	const uint particleID = cInput.GlobalThreadID.x;

	const uint count =  StreamInfo.StreamInfo [ 0 ] ;
	if (particleID >= count)
		return;

	const uint drawRequest =  Info.DrawRequest ;

	const uint positionsOffset =  PositionsOffsets.PositionsOffsets [ drawRequest ] ;
	const vec3 pos =  vec3( uintBitsToFloat(GPUSimData.GPUSimData [positionsOffset + particleID * 3 ]) , uintBitsToFloat(GPUSimData.GPUSimData [(positionsOffset + particleID * 3 ) + 1]) , uintBitsToFloat(GPUSimData.GPUSimData [(positionsOffset + particleID * 3 ) + 2]) ) ;

	const uint scalesOffset =  ScalesOffsets.ScalesOffsets [ drawRequest ] ;
	const vec3 scale =  vec3( uintBitsToFloat(GPUSimData.GPUSimData [scalesOffset + particleID * 3 ]) , uintBitsToFloat(GPUSimData.GPUSimData [(scalesOffset + particleID * 3 ) + 1]) , uintBitsToFloat(GPUSimData.GPUSimData [(scalesOffset + particleID * 3 ) + 2]) ) ;

	const uint orientationsOffset =  OrientationsOffsets.OrientationsOffsets [ drawRequest ] ;
	const vec4 q =  vec4( uintBitsToFloat(GPUSimData.GPUSimData [orientationsOffset + particleID * 4 ]) , uintBitsToFloat(GPUSimData.GPUSimData [(orientationsOffset + particleID * 4 ) + 1]) , uintBitsToFloat(GPUSimData.GPUSimData [(orientationsOffset + particleID * 4 ) + 2]) , uintBitsToFloat(GPUSimData.GPUSimData [(orientationsOffset + particleID * 4 ) + 3]) ) ;


	const vec3 r0 = vec3( 1 - 2 * (q.y * q.y + q.z * q.z),
								2 * (q.y * q.x + q.z * q.w),
								2 * (q.z * q.x - q.y * q.w));

	const vec3 r1 = vec3( 2 * (q.x * q.y - q.z * q.w),
						1 - 2 * (q.z * q.z + q.x * q.x),
							2 * (q.z * q.y + q.x * q.w));

	const vec3 r2 = vec3( 2 * (q.x * q.z + q.y * q.w),
							2 * (q.y * q.z - q.x * q.w),
						1 - 2 * (q.y * q.y + q.x * q.x));

	const vec4 m0 = vec4(r0 * scale.x, 0.0f);
	const vec4 m1 = vec4(r1 * scale.y, 0.0f);
	const vec4 m2 = vec4(r2 * scale.z, 0.0f);
	const vec4 m3 = vec4(pos, 1.0f);

	const uint matricesOffset =  MatricesOffsets.MatricesOffsets [ drawRequest ] ;

	OutMatrices.OutMatrices [matricesOffset + particleID * 16 + 4 * 0 ] = floatBitsToUint(m0.x) ; OutMatrices.OutMatrices [(matricesOffset + particleID * 16 + 4 * 0 ) + 1] = floatBitsToUint(m0.y) ; OutMatrices.OutMatrices [(matricesOffset + particleID * 16 + 4 * 0 ) + 2] = floatBitsToUint(m0.z) ; OutMatrices.OutMatrices [(matricesOffset + particleID * 16 + 4 * 0 ) + 3] = floatBitsToUint(m0.w) ; ;
	OutMatrices.OutMatrices [matricesOffset + particleID * 16 + 4 * 1 ] = floatBitsToUint(m1.x) ; OutMatrices.OutMatrices [(matricesOffset + particleID * 16 + 4 * 1 ) + 1] = floatBitsToUint(m1.y) ; OutMatrices.OutMatrices [(matricesOffset + particleID * 16 + 4 * 1 ) + 2] = floatBitsToUint(m1.z) ; OutMatrices.OutMatrices [(matricesOffset + particleID * 16 + 4 * 1 ) + 3] = floatBitsToUint(m1.w) ; ;
	OutMatrices.OutMatrices [matricesOffset + particleID * 16 + 4 * 2 ] = floatBitsToUint(m2.x) ; OutMatrices.OutMatrices [(matricesOffset + particleID * 16 + 4 * 2 ) + 1] = floatBitsToUint(m2.y) ; OutMatrices.OutMatrices [(matricesOffset + particleID * 16 + 4 * 2 ) + 2] = floatBitsToUint(m2.z) ; OutMatrices.OutMatrices [(matricesOffset + particleID * 16 + 4 * 2 ) + 3] = floatBitsToUint(m2.w) ; ;
	OutMatrices.OutMatrices [matricesOffset + particleID * 16 + 4 * 3 ] = floatBitsToUint(m3.x) ; OutMatrices.OutMatrices [(matricesOffset + particleID * 16 + 4 * 3 ) + 1] = floatBitsToUint(m3.y) ; OutMatrices.OutMatrices [(matricesOffset + particleID * 16 + 4 * 3 ) + 2] = floatBitsToUint(m3.z) ; OutMatrices.OutMatrices [(matricesOffset + particleID * 16 + 4 * 3 ) + 3] = floatBitsToUint(m3.w) ; ;
}

layout(local_size_x = 128, local_size_y = 1, local_size_z = 1) in;
void main()
{
	SComputeInput cInput;
	cInput.GlobalThreadID = gl_GlobalInvocationID;
	cInput.LocalThreadID = gl_LocalInvocationID;
	cInput.GroupID = gl_WorkGroupID;
	ComputeMain(cInput);
}
